const fs = require('fs');
const { join } = require('path');

const defaultPolyfills = ['es.array.iterator', 'es.promise', 'es.object.assign'];

/**
 *
 * @param {import('./interfaces').Options['browserslist']} browserList
 * @param {'production' | 'development'} env
 */
function getBrowserList(browserList, env) {
  if (typeof browserList === 'object' && 'production' in browserList) {
    return browserList[env] || browserList.production;
  } else if (Array.isArray(browserList)) {
    return browserList;
  } else if (typeof browserList === 'string') {
    return browserList;
  }
}

/**
 *
 * @param {'production' | 'development'} env
 * @returns {undefined | string | string[]}
 */
function getBrowserListFromUserProject(env) {
  try {
    const cwd = process.cwd();
    const pkgJson = join(cwd, 'package.json');
    if (fs.existsSync(pkgJson)) {
      const pkg = require(pkgJson);
      const browserList = pkg.browserslist;
      return getBrowserList(browserList, env);
    }
  } catch (error) {
    error;
  }
}

const defaultDevelopmentBrowserList = ['last 5 chrome version', 'last 5 firefox version', 'last 1 safari version'];

const defaultProductionBrowserList = ['last 2 versions', 'Firefox ESR', 'not dead', 'not ie <= 10'];

/**
 *
 * @param {import('./interfaces').Options} options
 */
function getFoxpageComponentBabelConfig(options = {}) {
  const {
    by,
    env,
    mode: buildMode,
    debug = false,
    developIE = false,
    esModule = false,
    loose = false,
    spec = false,
    modules,
    presetEnvOptions: presetEnv = {},
    style = false,
    browserslist,
  } = options;

  const isBuildByWebpack = by === 'webpack';
  const isBuildProduction = buildMode === 'production';
  const isBuildDebug = buildMode === 'debug';
  const isBuildNode = buildMode === 'node';
  const isBuildEditor = buildMode === 'editor';
  const isStorybook = process.env.STORYBOOK === 'true';

  const isProduction = env === 'production';
  const isDevelopment = env === 'development';
  const isTest = env === 'test';
  const browserListFromPackageJson = getBrowserListFromUserProject(env);
  const browserListFromOption = getBrowserList(browserslist, env);

  const defaultModules = esModule ? false : isBuildByWebpack ? false : 'auto';

  const presets = [];
  const plugins = [];

  const presetEnvCommonOptions = {
    corejs: { version: 3, proposals: true },
    useBuiltIns: 'usage',
    debug,
    loose,
    spec,
    modules: typeof modules !== 'undefined' ? modules : defaultModules,
  };

  const prodEnvOptions = {
    ...presetEnvCommonOptions,
    targets: {
      browsers: browserListFromOption || browserListFromPackageJson || defaultProductionBrowserList,
    },
    ...presetEnv,
  };

  const debugEnvOptions = {
    ...prodEnvOptions,
  };

  const devEnvOptions = {
    ...presetEnvCommonOptions,
    loose: true,
    targets: developIE
      ? { ie: 11 }
      : { browsers: browserListFromOption || browserListFromPackageJson || defaultDevelopmentBrowserList },
    ...presetEnv,
  };

  const nodeEnvOptions = {
    ...presetEnvCommonOptions,
    loose: true,
    targets: {
      node: 'current',
    },
    ...presetEnv,
  };

  const editorEnvOption = {
    ...presetEnvCommonOptions,
    loose: true,
    targets: {
      browsers: 'last 5 Chrome versions',
    },
    ...presetEnv,
  };

  const mapModeToEnvOption = {
    production: prodEnvOptions,
    debug: debugEnvOptions,
    node: nodeEnvOptions,
    editor: editorEnvOption,
  };

  const presetEnvOptions = mapModeToEnvOption[buildMode] || (isDevelopment ? devEnvOptions : prodEnvOptions);

  if (debug) {
    console.log('babel preset-env options:', presetEnvOptions);
  }

  presets.push(
    [
      require('@babel/preset-typescript'),
      {
        isTSX: true,
        allExtensions: true,
        allowNamespaces: true,
      },
    ],
    [
      require('@babel/preset-react'),
      {
        runtime: 'automatic',
        development: !isProduction || isStorybook,
      },
    ],
    [require('@babel/preset-env'), presetEnvOptions],
  );

  plugins.push(
    [require('@babel/plugin-proposal-private-methods'), { loose }],
    [require('@babel/plugin-proposal-class-properties'), { loose }],
    require('@babel/plugin-proposal-export-namespace-from'),
    require('@babel/plugin-proposal-export-default-from'),
    require('@babel/plugin-proposal-numeric-separator'),
    require('@babel/plugin-syntax-dynamic-import'),
    [
      require('@babel/plugin-transform-runtime'),
      {
        corejs: 3,
        helpers: true,
        regenerator: true,
      },
    ],
    [require('@babel/plugin-proposal-optional-chaining'), { loose }],
    [require('@babel/plugin-proposal-nullish-coalescing-operator'), { loose }],
  );

  if (isTest) {
    presets.push(require('babel-preset-power-assert'));
  }

  if (isBuildByWebpack && (isProduction || isBuildProduction) && !isStorybook) {
    plugins.push(require('babel-plugin-transform-react-remove-prop-types'));
    plugins.push(require('babel-plugin-transform-remove-console'));
  } else if (isBuildNode) {
    plugins.push(require('babel-plugin-transform-remove-console'));
  }

  if (style) {
    const transformStylePluginOption = style === 'remove' ? { removeStyleImport: true } : undefined;
    if (transformStylePluginOption) {
      plugins.push([require('./plugins/transform-style-import'), transformStylePluginOption]);
    }
  }

  return {
    presets,
    plugins,
  };
}

module.exports = {
  getFoxpageComponentBabelConfig,
};
